%{
(*
 * pl0lex.l
 *
 * lex input file for PL/0 scanner
 *
 *   adapted to Turbo Pascal Lex  HW
 *   Note: keywords are determined by scanning a keyword table, rather
 *   than including the keyword patterns in the Lex source which is done
 *   in the original version of this file. The main reason for this is,
 *   that the DFA table generated from the original spec grows too large
 *   to be compilable by Turbo Pascal (Lex will construct the table,
 *   but Turbo Pascal puts it into the static data segment which holds
 *   all typed constants and global variables, and this data segment is
 *   64 KB max.).
 *)

UNIT PL0Lex;

INTERFACE

USES PL0INIT, LexLib, YaccLib;

{$I PL0.H}

VAR
  line_no : integer;

function yylex : integer;
function act_token : string;

IMPLEMENTATION

const newline = #10;

procedure commenteof;
  begin
    writeln('unexpected EOF inside comment at line ', line_no);
  end(*commenteof*);

function is_keyword(id : string; var token : integer) : boolean; 
  (* checks whether id is PL0 keyword; if so, returns corresponding
     token number in token *)
  forward;
%}

%%

 var
   c  : char;
   kw : integer;
   cd : integer;

[a-zA-Z]([a-zA-Z0-9])*	if is_keyword(yytext, kw) then
                          return(kw)
                        else
                        begin
                          yylval.ident:= yytext;
                          return(IDENTIFIER);
                        end;

"'"[^']*"'"             begin
                          yylval.ident:=
                             copy(yytext, 2, length(yytext)-2);
                          return(_STRING);
                        end;

":="                    return(ASSIGNMENT);
":"			return(COLON);
","			return(COMMA);
[0-9]+                  begin
                           val(yytext, yylval.number, cd);
                           return(NUMBER);
                        end;

"."			return(DOT);
"="			return(EQUAL);
">="                    return(GREATERE);
">"                     return(GREATER);
"<="                    return(LESSE);
"("			return(LPAREN);
"<"                     return(LESS);
"-"			return(MINUS);
"#"                     return(NOTEQUAL);
"+"			return(PLUS);
")"			return(RPAREN);
";"			return(SEMICOLON);
"/"			return(SLASH);
"*"			return(STAR);
"["                     return(LBRACKET);
"]"                     return(RBRACKET);
"(*"                    begin
                          repeat
			    c := input;
			    case c of
			      '*' : begin
				      c := input;
				      if c=')' then exit else unput(c)
				    end;
			      newline : inc(line_no);
                              #0 : commenteof;
			    end;
			  until false
                        end;

[ \t\f]                 ;

\n                      inc(line_no);

.			writeln('''', yytext[1], ''' (#', ord(yytext[1]),
			  '): illegal character at line ', line_no);

%%

function upper(id : string) : string;
  var i : integer;
  begin
    for i := 1 to length(id) do
      id[i] := upCase(id[i]);
    upper := id
  end(*upper*);

function is_keyword(id : string; var token : integer) : boolean;
  const
    id_len = 20;
  type
    Ident = string[id_len];
  const
    (* table of PL0 keywords: *)
    no_of_keywords = 22;
    keyword : array [1..no_of_keywords] of Ident =
      (
      'BEGIN',
      'CALL',
      'CASE',
      'CONST',
      'DO',
      'DOWNTO',
      'ELSE',
      'END',
      'FOR',
      'IF',
      'ODD',
      'OF',
      'PROCEDURE',
      'READ',
      'REPEAT',
      'THEN',
      'TO',
      'UNTIL',
      'VAR',
      'WHILE',
      'WRITE',
      'WRITELN'
      );
    keyword_token : array [1..no_of_keywords] of integer =
      (
      _BEGIN,
      CALL,
      _CASE,
      _CONST,
      _DO,
      _DOWNTO,
      _ELSE,
      _END,
      _FOR,
      _IF,
      ODD,
      _OF,
      _PROCEDURE,
      _READ,
      _REPEAT,
      _THEN,
      _TO,
      _UNTIL,
      _VAR,
      _WHILE,
      _WRITE,
      _WRITELN
      );

  var m, n, k : integer;
  begin
    id := upper(id);
    m := 1; n := no_of_keywords;
    while m<=n do
      begin
        k := m+(n-m) div 2;
        if id=keyword[k] then
          begin
            is_keyword := true;
            token := keyword_token[k];
            exit
          end
        else if id>keyword[k] then
          m := k+1
        else
          n := k-1
      end;
    is_keyword := false
  end(*is_keyword*);

function act_token : string;
  begin
    act_token := yytext
  end(*act_token*);

begin
  if paramcount = 0 then
  begin
    write('Input file [.PL0]: ');
    readln(sourcefile);
  end
  else
    sourcefile:= paramstr(1);
  if pos('.', sourcefile) = 0 then
     sourcefile:= sourcefile + '.pl0';
  line_no := 1;
  assign(yyin, sourcefile);
  (*$I-*) reset(yyin); (*$I+*)
  IF IOResult <> 0 THEN
  BEGIN
     WriteLn('File ', sourcefile, ' not found.');
     HALT(1)
  END;
end.
